"""
optimizelab - A collection of classical and nature-inspired optimization algorithms.

Author: code spaze
"""

__all__ = [
    "continuous_ga",
    "binary_ga",
    "simulated_annealing",
    "ant_colony",
    "particle_swarm",
    "gray_wolf",
    "tabu_search",
    "shuffled_frog",
    "tsp_genetic",
]


# Expose entire code as a variable
__init__ = "\"\"\"\noptimizelab - A collection of classical and nature-inspired optimization algorithms.\n\nAuthor: code spaze\n\"\"\"\n\n__all__ = [\n    \"continuous_ga\",\n    \"binary_ga\",\n    \"simulated_annealing\",\n    \"ant_colony\",\n    \"particle_swarm\",\n    \"gray_wolf\",\n    \"tabu_search\",\n    \"shuffled_frog\",\n    \"tsp_genetic\",\n]\n"
